@@ -5,6 +5,7 @@
 #include <string.h>
 #include <unistd.h>
 
+#include "blake2.h"
 #include "dedup.h"
 
 #define ROTL(x, y) (((x) << (y)) | ((x) >> (32 - (y))))
@@ -12,9 +13,12 @@
 struct chunker {
 	uint8_t *buf;
 	int fd;
-	size_t cap;
 	size_t rpos;
 	size_t wpos;
+	size_t min_size;
+	size_t max_size;
+	size_t mask;
+	size_t win_size;
 };
 
 /*
@@ -26,7 +30,7 @@ struct chunker {
  * exactly 50% chance that a XOR operation would flip all the bits in
  * the hash.
  */
-static uint32_t buz[] = {
+static const uint32_t buz[] = {
 	0xbc9fa594,0x30a8f827,0xced627a7,0xdb46a745,0xcfa4a9e8,0x77cccb59,0xddb66276,0x3adc532f,
 	0xfe8b67d3,0x8155b59e,0x0c893666,0x1d757009,0x17394ee4,0x85d94c07,0xcacd52da,0x076c6f79,
 	0xead0a798,0x6c7ccb4a,0x2639a1b8,0x3aa5ae32,0x3e6218d2,0xb290d980,0xa5149521,0x4b426119,
@@ -65,65 +69,65 @@ static uint32_t buz[] = {
 static inline uint32_t
 buzh_init(uint8_t *buf, size_t size)
 {
-	uint32_t fp;
+	uint32_t sum;
 	size_t i;
 
-	for (i = 1, fp = 0; i < size; i++, buf++)
-		fp ^= ROTL(buz[*buf], (size - i) % 32);
+	for (i = 1, sum = 0; i < size; i++, buf++)
+		sum ^= ROTL(buz[*buf], (size - i) % 32);
 
-	return fp ^ buz[*buf];
+	return sum ^ buz[*buf];
 }
 
 static inline uint32_t
-buzh_update(uint32_t fp, uint8_t out, uint8_t in, size_t size)
+buzh_update(uint32_t sum, uint8_t out, uint8_t in, size_t size)
 {
-	return ROTL(fp, 1) ^ ROTL(buz[out], size % 32) ^ buz[in];
-}
-
-static inline int
-match_pattern(size_t chunk_size, uint32_t fp)
-{
-	if (chunk_size >= BLKSIZE_MAX)
-		return 1;
-	if (chunk_size < BLKSIZE_MIN)
-		return 0;
-	return (fp & HASHMASK_BITS) == 0;
+	return ROTL(sum, 1) ^ ROTL(buz[out], size % 32) ^ buz[in];
 }
 
 static size_t
 get_chunk_size(struct chunker *chunker)
 {
+	size_t max_chunk_size, win_size, i;
+	uint32_t sum;
 	uint8_t *bp;
-	uint32_t fp;
-	size_t i, chunk_size;
 
-	chunk_size = chunker->wpos - chunker->rpos;
-	if (chunk_size < WINSIZE)
-		return chunk_size;
+	max_chunk_size = chunker->wpos - chunker->rpos;
+	win_size = chunker->win_size;
+	if (max_chunk_size < win_size)
+		return max_chunk_size;
 
 	/*
 	 * To achieve better deduplication, we chunk blocks based on a
-	 * recurring pattern occuring on the data stream. A fixed window
-	 * of WINSIZE bytes is slid over the data, and a rolling hash is
-	 * computed for this window.
+	 * recurring pattern on the data stream. We slide a fixed window
+	 * of WINSIZE bytes over the data, and a rolling hash is computed
+	 * for this window.
 	 * When the rolling hash matches a given pattern the block is chunked
 	 * at the end of that window.
 	 */
 	bp = &chunker->buf[chunker->rpos];
-	fp = buzh_init(bp, WINSIZE);
-	for (i = 0; i < chunk_size - WINSIZE; i++) {
-		if (i > 0)
-			fp = buzh_update(fp, bp[i - 1], bp[i + WINSIZE - 1],
-			                 WINSIZE);
-		if (match_pattern(i + WINSIZE, fp))
-			return i + WINSIZE;
-			
+	sum = buzh_init(bp, win_size);
+	for (i = 0; i < max_chunk_size - win_size; i++) {
+		size_t chunk_size = i + win_size;
+
+		if (i > 0) {
+			uint8_t out = bp[i - 1];
+			uint8_t in = bp[chunk_size - 1];
+
+			sum = buzh_update(sum, out, in, win_size);
+		}
+
+		if (chunk_size < chunker->min_size)
+			continue;
+
+		if ((sum & chunker->mask) == 0)
+			return chunk_size;
 	}
-	return chunk_size;
+	return max_chunk_size;
 }
 
 struct chunker *
-alloc_chunker(int fd, size_t cap)
+alloc_chunker(int fd, size_t min_size, size_t max_size,
+              size_t mask, size_t win_size)
 {
 	struct chunker *chunker;
 
@@ -131,11 +135,15 @@ alloc_chunker(int fd, size_t cap)
 	if (chunker == NULL)
 		err(1, "calloc");
 
-	chunker->buf = calloc(1, cap);
+	chunker->buf = calloc(1, max_size);
 	if (chunker->buf == NULL)
 		err(1, "calloc");
+
 	chunker->fd = fd;
-	chunker->cap = cap;
+	chunker->min_size = min_size;
+	chunker->max_size = max_size;
+	chunker->mask = mask;
+	chunker->win_size = win_size;
 
 	return chunker;
 }
@@ -154,7 +162,7 @@ fill_chunker(struct chunker *chunker)
 	ssize_t n;
 
 	bp = &chunker->buf[chunker->wpos];
-	n = xread(chunker->fd, bp, chunker->cap - chunker->wpos);
+	n = xread(chunker->fd, bp, chunker->max_size - chunker->wpos);
 	chunker->wpos += n;
 	return chunker->wpos;
 }
