@@ -30,11 +30,11 @@ struct extract_args {
 
 static struct snap_hdr snap_hdr;
 static struct blk_hdr blk_hdr;
-static struct cache *cache;
+static struct icache *icache;
 static int ifd;
 static int sfd;
-static unsigned long long cache_hits;
-static unsigned long long cache_misses;
+static int hash_algo = HASH_BLAKE2B;
+static int compr_algo = COMPR_LZ4;
 
 int verbose;
 char *argv0;
@@ -51,6 +51,9 @@ print_md(FILE *fp, uint8_t *md, size_t size)
 static void
 print_stats(struct stats *st)
 {
+	unsigned long long hits, misses;
+	double hitratio;
+
 	if (st->nr_blks == 0)
 		return;
 
@@ -68,8 +71,15 @@ print_stats(struct stats *st)
 	        (unsigned long long)st->max_blk_size);
 	fprintf(stderr, "Number of unique blocks: %llu\n",
 	        (unsigned long long)st->nr_blks);
-	fprintf(stderr, "Cache hits: %llu\n", cache_hits);
-	fprintf(stderr, "Cache misses: %llu\n", cache_misses);
+
+	icache_stats(icache, &hits, &misses);
+	if (hits == 0 && misses == 0)
+		hitratio = 0;
+	else
+		hitratio = (double)hits / (hits + misses);
+
+	fprintf(stderr, "Index cache hit percentage: %.2f%%\n",
+	        100 * hitratio);
 }
 
 static struct snap *
@@ -96,17 +106,18 @@ free_snap(struct snap *snap)
 static void
 hash_snap(struct snap *snap, uint8_t *md)
 {
-	blake2b_state ctx;
+	struct hash_ctx ctx;
 	uint64_t i;
 
-	blake2b_init(&ctx, MDSIZE);
+	if (hash_init(&ctx, hash_algo, MD_SIZE) < 0)
+		errx(1, "hash_init failed");
 	for (i = 0; i < snap->nr_blk_descs; i++) {
 		struct blk_desc *blk_desc;
 
 		blk_desc = &snap->blk_desc[i];
-		blake2b_update(&ctx, blk_desc->md, sizeof(blk_desc->md));
+		hash_update(&ctx, blk_desc->md, sizeof(blk_desc->md));
 	}
-	blake2b_final(&ctx, md, MDSIZE);
+	hash_final(&ctx, md, MD_SIZE);
 }
 
 static struct snap *
@@ -172,11 +183,12 @@ free_buf(uint8_t *buf)
 static void
 hash_blk(uint8_t *buf, size_t size, uint8_t *md)
 {
-	blake2b_state ctx;
+	struct hash_ctx ctx;
 
-	blake2b_init(&ctx, MDSIZE);
-	blake2b_update(&ctx, buf, size);
-	blake2b_final(&ctx, md, MDSIZE);
+	if (hash_init(&ctx, hash_algo, MD_SIZE) < 0)
+		errx(1, "hash_init failed");
+	hash_update(&ctx, buf, size);
+	hash_final(&ctx, md, MD_SIZE);
 }
 
 static void
@@ -206,29 +218,32 @@ append_blk(uint8_t *buf, struct blk_desc *blk_desc)
 static void
 dedup_chunk(struct snap *snap, uint8_t *chunkp, size_t chunk_size)
 {
-	uint8_t md[MDSIZE];
+	uint8_t md[MD_SIZE];
 	struct blk_desc blk_desc;
+	struct compr_ctx ctx;
 	uint8_t *compr_buf;
-	size_t n;
+	size_t n, csize;
 
-	compr_buf = alloc_buf(compr_size(BLKSIZE_MAX));
+	if (compr_init(&ctx, compr_algo) < 0)
+		errx(1, "compr_init failed");
+	csize = compr_size(&ctx, BLKSIZE_MAX);
+	compr_buf = alloc_buf(csize);
 
-	n = compr(chunkp, compr_buf, chunk_size, compr_size(BLKSIZE_MAX));
+	n = compr(&ctx, chunkp, compr_buf, chunk_size, csize);
 	hash_blk(compr_buf, n, md);
 
 	snap_hdr.st.orig_size += chunk_size;
 	snap_hdr.st.compr_size += n;
 
 	memcpy(blk_desc.md, md, sizeof(blk_desc.md));
-	if (lookup_cache_entry(cache, &blk_desc) < 0) {
+	if (lookup_icache(icache, &blk_desc) < 0) {
 		blk_desc.offset = blk_hdr.size;
 		blk_desc.size = n;
 
 		snap->blk_desc[snap->nr_blk_descs++] = blk_desc;
 		append_blk(compr_buf, &blk_desc);
 
-		add_cache_entry(cache, &blk_desc);
-		cache_misses++;
+		insert_icache(icache, &blk_desc);
 
 		snap_hdr.st.dedup_size += blk_desc.size;
 		snap_hdr.st.nr_blks++;
@@ -239,10 +254,10 @@ dedup_chunk(struct snap *snap, uint8_t *chunkp, size_t chunk_size)
 			snap_hdr.st.min_blk_size = blk_desc.size;
 	} else {
 		snap->blk_desc[snap->nr_blk_descs++] = blk_desc;
-		cache_hits++;
 	}
 
 	free(compr_buf);
+	compr_final(&ctx);
 }
 
 static void
@@ -288,24 +303,28 @@ extract(struct snap *snap, void *arg)
 {
 	uint8_t *buf[2];
 	struct extract_args *args = arg;
+	struct compr_ctx ctx;
 	uint64_t i;
 
 	if (memcmp(snap->md, args->md, sizeof(snap->md)) != 0)
 		return WALK_CONTINUE;
 
+	if (compr_init(&ctx, compr_algo) < 0)
+		errx(1, "compr_init failed");
 	buf[0] = alloc_buf(BLKSIZE_MAX);
-	buf[1] = alloc_buf(compr_size(BLKSIZE_MAX));
+	buf[1] = alloc_buf(compr_size(&ctx, BLKSIZE_MAX));
 	for (i = 0; i < snap->nr_blk_descs; i++) {
 		struct blk_desc *blk_desc;
 		size_t blksize;
 
 		blk_desc = &snap->blk_desc[i];
 		read_blk(buf[1], blk_desc);
-		blksize = decompr(buf[1], buf[0], blk_desc->size, BLKSIZE_MAX);
+		blksize = decompr(&ctx, buf[1], buf[0], blk_desc->size, BLKSIZE_MAX);
 		xwrite(args->fd, buf[0], blksize);
 	}
 	free_buf(buf[1]);
 	free_buf(buf[0]);
+	compr_final(&ctx);
 	args->ret = 0;
 	return WALK_STOP;
 }
@@ -318,6 +337,7 @@ extract(struct snap *snap, void *arg)
 static int
 check_snap(struct snap *snap, void *arg)
 {
+	struct compr_ctx ctx;
 	uint8_t *buf;
 	int *ret = arg;
 	uint64_t i;
@@ -328,18 +348,16 @@ check_snap(struct snap *snap, void *arg)
 		fputc('\n', stderr);
 	}
 
-	buf = alloc_buf(compr_size(BLKSIZE_MAX));
+	if (compr_init(&ctx, compr_algo) < 0)
+		errx(1, "compr_init failed");
+	buf = alloc_buf(compr_size(&ctx, BLKSIZE_MAX));
 	for (i = 0; i < snap->nr_blk_descs; i++) {
-		uint8_t md[MDSIZE];
-		blake2b_state ctx;
+		uint8_t md[MD_SIZE];
 		struct blk_desc *blk_desc;
 
 		blk_desc = &snap->blk_desc[i];
 		read_blk(buf, blk_desc);
-
-		blake2b_init(&ctx, MDSIZE);
-		blake2b_update(&ctx, buf, blk_desc->size);
-		blake2b_final(&ctx, md, MDSIZE);
+		hash_blk(buf, blk_desc->size, md);
 
 		if (memcmp(blk_desc->md, md, sizeof(blk_desc->md)) == 0)
 			continue;
@@ -358,23 +376,28 @@ check_snap(struct snap *snap, void *arg)
 		*ret = -1;
 	}
 	free_buf(buf);
+	compr_final(&ctx);
 	return WALK_CONTINUE;
 }
 
 static int
-build_cache(struct snap *snap, void *arg)
+build_icache(struct snap *snap, void *arg)
 {
+	struct compr_ctx ctx;
 	uint8_t *buf;
 	uint64_t i;
 
-	buf = alloc_buf(compr_size(BLKSIZE_MAX));
+	if (compr_init(&ctx, compr_algo) < 0)
+		errx(1, "compr_init failed");
+	buf = alloc_buf(compr_size(&ctx, BLKSIZE_MAX));
 	for (i = 0; i < snap->nr_blk_descs; i++) {
 		struct blk_desc *blk_desc;
 
 		blk_desc = &snap->blk_desc[i];
-		add_cache_entry(cache, blk_desc);
+		insert_icache(icache, blk_desc);
 	}
 	free(buf);
+	compr_final(&ctx);
 	return WALK_CONTINUE;
 }
 
@@ -429,7 +452,8 @@ static void
 init_blk_hdr(void)
 {
 	blk_hdr.flags = (VER_MAJ << VER_MAJ_SHIFT) | VER_MIN;
-	blk_hdr.flags |= compr_enabled << COMPR_ENABLED_SHIFT;
+	blk_hdr.flags |= compr_algo << COMPR_ALGO_SHIFT;
+	blk_hdr.flags |= hash_algo << HASH_ALGO_SHIFT;
 	blk_hdr.size = BLK_HDR_SIZE;
 }
 
@@ -442,9 +466,27 @@ load_blk_hdr(void)
 	read_blk_hdr(sfd, &blk_hdr);
 	match_ver(blk_hdr.flags);
 
-	v = blk_hdr.flags >> COMPR_ENABLED_SHIFT;
-	v &= COMPR_ENABLED_MASK;
-	compr_enabled = v;
+	v = blk_hdr.flags >> COMPR_ALGO_SHIFT;
+	v &= COMPR_ALGO_MASK;
+	compr_algo = v;
+
+	if (compr_algo < 0 || compr_algo >= NR_COMPRS)
+		errx(1, "unsupported compression algorithm: %d", compr_algo);
+
+	if (verbose > 0)
+		fprintf(stderr, "Compression algorithm: %s\n",
+		        compr_type2name(compr_algo));
+
+	v = blk_hdr.flags >> HASH_ALGO_SHIFT;
+	v &= HASH_ALGO_MASK;
+	hash_algo = v;
+
+	if (hash_algo < 0 || hash_algo >= NR_HASHES)
+		errx(1, "unsupported hash algorithm: %d", hash_algo);
+
+	if (verbose > 0)
+		fprintf(stderr, "Hash algorithm: %s\n",
+		        hash_type2name(hash_algo));
 }
 
 static void
@@ -457,9 +499,14 @@ save_blk_hdr(void)
 static void
 init_snap_hdr(void)
 {
+	struct compr_ctx ctx;
+
+	if (compr_init(&ctx, compr_algo) < 0)
+		errx(1, "compr_init failed");
 	snap_hdr.flags = (VER_MAJ << VER_MAJ_SHIFT) | VER_MIN;
 	snap_hdr.size = SNAP_HDR_SIZE;
-	snap_hdr.st.min_blk_size = compr_size(BLKSIZE_MAX);
+	snap_hdr.st.min_blk_size = compr_size(&ctx, BLKSIZE_MAX);
+	compr_final(&ctx);
 }
 
 static void
@@ -506,8 +553,8 @@ init(int iflag)
 		load_blk_hdr();
 	}
 
-	cache = alloc_cache();
-	walk_snap(build_cache, NULL);
+	icache = alloc_icache();
+	walk_snap(build_icache, NULL);
 }
 
 static void
@@ -516,7 +563,7 @@ term(void)
 	if (verbose > 0)
 		print_stats(&snap_hdr.st);
 
-	free_cache(cache);
+	free_icache(icache);
 
 	save_blk_hdr();
 	save_snap_hdr();
@@ -531,21 +578,31 @@ term(void)
 static void
 usage(void)
 {
-	fprintf(stderr, "usage: %s [-Zcilv] [-e id] [-r root] [-m message] [file]\n", argv0);
+	fprintf(stderr, "usage: %s [-cilv] [-H hash] [-Z compressor] [-e id] [-r root] [-m message] [file]\n", argv0);
 	exit(1);
 }
 
 int
 main(int argc, char *argv[])
 {
-	uint8_t md[MDSIZE];
-	char *id = NULL, *root = NULL, *msg = NULL;
+	uint8_t md[MD_SIZE];
+	char *id = NULL, *root = NULL, *msg = NULL, *hash_name = NULL;
+	char *compr_name;
 	int iflag = 0, lflag = 0, cflag = 0;
 	int fd = -1;
 
 	ARGBEGIN {
+	case 'H':
+		hash_name = EARGF(usage());
+		hash_algo = hash_name2type(hash_name);
+		if (hash_algo < 0)
+			errx(1, "unknown hash: %s", hash_name);
+		break;
 	case 'Z':
-		compr_enabled = 0;
+		compr_name = EARGF(usage());
+		compr_algo = compr_name2type(compr_name);
+		if (compr_algo < 0)
+			errx(1, "unknown compressor: %s", compr_name);
 		break;
 	case 'c':
 		cflag = 1;
