@@ -6,6 +6,7 @@
 #include <stdint.h>
 #include <stdlib.h>
 
+#include "blake2.h"
 #include "dedup.h"
 
 void
@@ -108,7 +109,7 @@ read_blk_desc(int fd, struct blk_desc *desc)
 	if (xread(fd, buf, sizeof(buf)) == 0)
 		errx(1, "%s: unexpected EOF", __func__);
 
-	snprintf(fmt, sizeof(fmt), "'%dqq", MDSIZE);
+	snprintf(fmt, sizeof(fmt), "'%dqq", MD_SIZE);
 	n = unpack(buf, fmt,
 	           desc->md,
 	           &desc->offset,
@@ -124,7 +125,7 @@ write_blk_desc(int fd, struct blk_desc *desc)
 	char fmt[BUFSIZ];
 	int n;
 
-	snprintf(fmt, sizeof(fmt), "'%dqq", MDSIZE);
+	snprintf(fmt, sizeof(fmt), "'%dqq", MD_SIZE);
 	n = pack(buf, fmt,
 	         desc->md,
 	         desc->offset,
@@ -144,7 +145,7 @@ read_snap(int fd, struct snap *snap)
 	if (xread(fd, buf, sizeof(buf)) == 0)
 		errx(1, "%s: unexpected EOF", __func__);
 
-	snprintf(fmt, sizeof(fmt), "q'%d'%dq", MSGSIZE, MDSIZE);
+	snprintf(fmt, sizeof(fmt), "q'%d'%dq", MSG_SIZE, MD_SIZE);
 	n = unpack(buf, fmt,
 	           &snap->size,
 	           snap->msg,
@@ -170,7 +171,7 @@ write_snap(int fd, struct snap *snap)
 	char fmt[BUFSIZ];
 	int n;
 
-	snprintf(fmt, sizeof(fmt), "q'%d'%dq", MSGSIZE, MDSIZE);
+	snprintf(fmt, sizeof(fmt), "q'%d'%dq", MSG_SIZE, MD_SIZE);
 	n = pack(buf, fmt,
 	         snap->size,
 	         snap->msg,
